%------------------------------------------------------------------------------
\section{BPL language}
%------------------------------------------------------------------------------

\paragraph{Usage}

\begin{verbatim}
ghci -Wall -package rh AbstractSyntax.lhs
\end{verbatim}

%if False
%- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
\subsection{Module header}
%- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 

> module AbstractSyntax
> where
> import Control.Monad.Reader
>
> import PrettierNG hiding (break)
> import Message hiding (break)

%endif

%- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
\subsection{Abstract syntax: declarations}
%- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 

> data Decl
>   =  Val Pat Expr
>   |  Function [(LVar, [Formal], Type, Expr)]
>   |  Type LTyCon [Formal] Type
>   |  Datatype LTyCon [Formal] [Alt]
>   |  Data LTyCon                                           -- Datentyp ohne Konstruktoren
>   |  Constr LCon Type
>   |  Exception Alt
>   |  Module LStrVar [StrFormal] StrExpr                    -- module/structure
>   |  Open StrExpr
>   |  Local [Decl] [Decl]
>   |  Interface LSigVar [[LTyVar]] SigExpr
>      deriving (Show)

> type Alt
>   =  (LCon, Maybe Type)

> instance Pretty Decl

%- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
\subsection{Abstract syntax: expressions}
%- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 

> type LVar  =  (Loc, String)
> type LCon  =  (Loc, String)

> type Var   =  String
> type Con   =  String

> data Expr  
>   =  Bool     Loc Bool
>   |  If       Expr Expr Expr                  -- alternative
>   |  UnOp     UnOp Expr                       -- unary operator
>   |  BinOp    Expr BinOp Expr                 -- binary operator
>   |  Nat      Loc Integer                     -- numeral
>   |  String   Loc String                      -- string literal
>   |  Var      LVar                            -- variable
>   |  Let      [Decl] Expr                     -- local declaration
>   |  Fun      [Formal] Expr                   -- abstraction
>   |  RecFun   LVar [Formal] Type Expr         -- recursive abstraction
>   |  App      Expr [Actual]                   -- application
>   |  Rec      Pat Expr                        -- recursion
>   |  Con      LCon                            -- constructor
>   |  Sel      Expr Lab                        -- record selection
>   |  Record   [Row Expr]                      -- record construction
>   |  Seq      Expr Expr                       -- sequence
>   |  Ref      Expr                            -- allocation
>   |  Deref    Expr                            -- dereferencing
>   |  Assign   Expr Expr                       -- assignment
>   |  While    Expr Expr                       -- while-loop
>   |  Case     Expr [(Pat, Expr)]              -- case analysis
>   |  Throw    Expr                            -- throw an exception
>   |  Try      Expr [(Pat, Expr)]              -- catch an exception
>   |  Continue Expr Expr                       -- apply a continuation
>   |  LetCC    LVar Type Expr                  -- let current continuation

>--   |  Call     Prim Expr

>   |  QualVar  StrExpr LVar                    -- qualified variable
>   |  QualCon  StrExpr LCon                    -- qualified constructor
>   |  Class    [Slot]                          -- class expression
>   |  New      Expr                            -- object creation

>   |  ArrFun   Expr LVar Expr                  -- array creation via function
>   |  ArrList  [Expr]                          -- array creation from list
>   |  Subscript Expr Expr                      -- array subscription
>   |  Typed    Expr Type                       -- explicitly typed expression
>      deriving (Show)

> instance Pretty Expr

> app e []  =  e
> app e es  =  App e es

> record [Field Nothing e] = e
> record fs                = Record fs


> data Actual
>   =  ActExpr Expr
>   |  ActType [Type]
>      deriving (Show)

> data Formal
>   =  ForPat  Pat
>   |  ForTypePat [LTyVar]
>      deriving (Show)

> data Row a
>   =  Field (Maybe Lab) a
>      deriving (Show)

> instance Pretty a => Pretty (Row a)

> ppRow pp (TyField Nothing Nothing a)   =  pp a
> ppRow pp (TyField (Just s) Nothing a)  =  block 6 (ppLab s <> text " = " <> pp a)

> data UnOp  
>   =  Neg -- Not??
>      deriving (Show)

> instance Pretty UnOp where
>   pretty Neg   = text "-"

> data BinOp 
>   =  Add | Sub | Mul | Div | Mod | Lt | Leq | Eq | Neq | Geq | Gt | And | Or | Cat
>      deriving (Show, Eq)

> instance Pretty BinOp where
>   pretty Add  =  text "+"
>   pretty Sub  =  text "-"
>   pretty Mul  =  text "*"
>   pretty Div  =  text "/"
>   pretty Mod  =  text "%"
>   pretty Lt   =  text "<"
>   pretty Leq  =  text "=<"
>   pretty Eq   =  text "=="
>   pretty Neq  =  text "/="
>   pretty Geq  =  text ">="
>   pretty Gt   =  text ">"
>   pretty And  =  text "&&"
>   pretty Or   =  text "||"
>   pretty Cat  =  text "^"

> data Prim
>   =  Chr | GetChar | Head | Null | Ord | PutChar | Show | Tail | Write | ReadLine | Size |
>      ReadFromFile | WriteToFile
>      deriving (Show)

> instance Pretty Prim where
>   pretty Chr      =  text "chr"
>   pretty GetChar  =  text "getchar"
>   pretty Head     =  text "head"
>   pretty Null     =  text "null"
>   pretty Ord      =  text "ord"
>   pretty PutChar  =  text "putchar"
>   pretty Show     =  text "show"
>   pretty Tail     =  text "tail"
>   pretty Write    =  text "write"
>   pretty ReadLine =  text "read-line"
>   pretty Size     =  text "size"
>   pretty ReadFromFile     =  text "read-from-file"
>   pretty WriteToFile     =  text "write-to-file"

Slots aka class declarations.

> data Slot
>   =  SlMethod LVar [Formal] Type Expr
>   |  SlLocal [Decl] [Slot]
>   |  SlInherit Expr
>   |  SlInheritIn LVar Expr [Slot]
>      deriving (Show)

%- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
\subsection{Abstract syntax: patterns}
%- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 

> data Pat
>   =  PatSucceed
>   |  PatVar     LVar
>   |  PatQualCon StrExpr LCon (Maybe Pat)
>   |  PatCon     LCon (Maybe Pat)                       -- constructor application
>   |  PatTyped   Pat Type
>   |  PatRecord  [Row Pat]
>   |  PatAlt     Pat Pat
>   |  PatBoth    Pat Pat
>      deriving (Show)

> patRecord [Field Nothing p] = p
> patRecord fs                = PatRecord fs

> instance Pretty Pat

> ppPat (PatTyped (PatVar (_,"_")) t1)  -- special construction in function types
>   =  ppType t1
> ppPat (PatVar x)
>   =  return (ltext x)

%- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
\subsection{Abstract syntax: types}
%- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 

> type LTyVar  =  (Loc, String)
> type LTyCon  =  (Loc, String)

> type TyVar   =  String
> type TyCon   =  String

> data Type
>   =  TyVar LTyVar
>   |  TyCon LTyCon
>   |  QualTyCon  StrExpr LTyCon
>   |  TyApp Type [Actual]
>   |  TyFun [Formal] Type
>   |  TyForall [Formal] Type
>   |  TyRecord [TyRow] (Maybe Int) -- optional row variable for type checking
>   |  TyPred LVar Type Expr
>   |  TyClass  [TySlot] (Maybe Int) -- optional slot variable for type checking
>   |  TyObject [TySlot] (Maybe Int) -- optional slot variable for type checking
>   -- for type checking
>   |  TVar Int -- unification variable
>   |  TNatString Int -- for builtin operators
>      deriving (Show)

> tyRecord [TyField Nothing Nothing t] = t
> tyRecord fs                          = TyRecord fs Nothing

Hmm. (x : t) ist nicht zulssig, da wir keine 1-Tupel
haben.

> data TyRow
>   =  TyField (Maybe Lab) (Maybe LVar) Type
>      deriving (Show)

> data TySlot
>   =  TyMethod LVar Type
>      deriving (Show)

> instance Pretty Type

> ltext (l,x) = text x

> type P a = Reader PState a
> data PState = PState
>                 {  pprec  :: !Int
>                 }

> withPrec :: Int -> P a -> P a
> withPrec n = local (\s -> s { pprec = n })

> parensPrec :: Int -> P Doc -> P Doc
> parensPrec m p = ask >>= \ s -> if pprec s > m then liftM parens p else p

Operator precedence (from parser):
right 1 "->"
nonassoc 0 FORALL
right 1 "&"
right 0 "|"
right 2 "||"
nonassoc 1 ":="
right 0 ";"

> pType :: Type -> Doc
> pType t = runReader (ppType t) (PState 0)

> parensType :: Type -> Doc
> parensType t = runReader (ppType t) (PState 11)

> ppType :: Type -> P Doc
> ppType (TyVar x)
>   =  return (ppTyName x)
> ppType (TyCon x)
>   =  return (ppTyName x)
> -- ppType (QualTyCon s x)
> --   =  ppStrExpr s <> text "." <> text x
> ppType (TyApp t [])
>   =  ppType t
> ppType (TyApp t ts)
>   =  parensPrec 10 $ do
>        pt  <- withPrec 10 $ ppType t
>        pts <- withPrec 11 $ mapM ppActual ts 
>        return (group (strip (pt : pts)))
> ppType (TyFun [] t)
>   =  ppType t
> {-
> ppType (TyFun ts t)
>   =  parens (group (strip (map ppFormal ts)  <> text " ->" <+> ppType t))
> -}
> -- Ralf suggests to pretty-print all functions with multiple arguments in a
> -- curried style:
> ppType (TyFun (t:ts) t')
>   =  parensPrec 1 $ do
>        pt  <- withPrec 2 $ ppFormal t
>        pt' <- withPrec 1 $ ppType (TyFun ts t')
>        return (group (pt <> text " " <> block 2 (text "->" <+> pt')))
> ppType (TyForall [] t)
>   =  ppType t

NEU:

 ppType (TyForall ts t)
   =  parens (group (text "forall " <> strip (map ppFormal ts) <> text " =>" <+> ppType t))

> ppType (TyForall (t:ts) t')
>   =  parensPrec 1 $ do
>        pt  <- withPrec 2 $ ppFormal t
>        pt' <- withPrec 1 $ ppType (TyForall ts t')
>        return (group (pt <> text " " <> block 2 (text "->" <+> pt')))
> ppType (TyRecord rs Nothing)
>   =  return (parens (group (intersperse (char ',' <> sp) (map (ppRow pType) rs))))  -- preliminary
> ppType (TyRecord rs (Just n))
>   =  return (parens (group (intersperse (char ',' <> sp) (map (ppRow pType) rs ++ [text ("_" ++ show n)]))))  -- preliminary
> ppType (TyClass ss Nothing)
>   =  do
>        pss <- withPrec 0 $ mapM (\ (TyMethod x t) -> ppType t >>= \ pt -> return $ text "method " <> block 5 (ltext x <> text " :" <+> pt)) ss
>        return (group (nest 3 (text "class" <> strip (empty : pss)) <+> text "end"))
> ppType (TyClass ss (Just n))
>   =  do
>        pss <- withPrec 0 $ mapM (\ (TyMethod x t) -> ppType t >>= \ pt -> return $ text "method " <> block 5 (ltext x <> text " :" <+> pt)) ss
>        return (group (nest 3 (text "class" <> strip (empty : pss ++ [text ("_" ++ show n)])) <+> text "end"))
> ppType (TyObject ss Nothing)
>   =  do
>        pss <- withPrec 0 $ mapM (\ (TyMethod x t) -> ppType t >>= \ pt -> return $ text "method " <> block 5 (ltext x <> text " :" <+> pt)) ss
>        return (group (nest 3 (text "object" <> strip (empty : pss)) <+> text "end"))
> ppType (TyObject ss (Just n))
>   =  do
>        pss <- withPrec 0 $ mapM (\ (TyMethod x t) -> ppType t >>= \ pt -> return $ text "method " <> block 5 (ltext x <> text " :" <+> pt)) ss
>        return (group (nest 3 (text "object" <> strip (empty : pss ++ [text ("_" ++ show n)])) <+> text "end"))
> ppType (TVar n)
>   =  return $ text ("_" ++ show n)
> ppType (TNatString n)
>   =  return $ text ("_ns" ++ show n)

> ppActual (ActExpr e)
>   =  return (text "<expression>") -- preliminary
> ppActual (ActType ts)
>   =  do
>        pts <- withPrec 0 $ mapM ppType ts
>        return (tbrackets (strip pts))

> pFormal f = runReader (ppFormal f) (PState 0)

> ppFormal (ForPat p)
>   =  ppPat p
> ppFormal (ForTypePat xs)
> --  =  tbrackets (intersperse (text ", ") (map ltext xs))
>   =  return (tbrackets (intersperse (text ", ") [ ppTyName x <> text " :" <+> text "Type" | x <- xs ]))

> tbrackets d = block 1 (string "<" <> d <> string ">")

%- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
\subsection{Abstract syntax: modules}
%- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 

> type LStrVar  =  (Loc,String)
> type StrVar   =  String

> data StrExpr
>   =  StrVar LStrVar
>   |  StrSel StrExpr LStrVar
>   |  Struct [Decl]
>   |  Functor [StrFormal] StrExpr
>   |  StrApp StrExpr [StrActual]
>      deriving (Show)

> data StrActual
>   =  StrActStrExpr StrExpr
>   |  StrActType [Type]
>      deriving (Show)

> data StrFormal
>   =  StrForStrPat LStrVar SigExpr
>   |  StrForTypePat [LTyVar]
>      deriving (Show)

> instance Pretty StrExpr

%- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
\subsection{Abstract syntax: interfaces}
%- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 

> type LSigVar  =  (Loc,String)
> type SigVar   =  String

> data SigExpr
>   =  SigVar LSigVar
>   |  SigSel StrExpr LSigVar
>   |  Sig [Spec]
>   |  SigFunctor [StrFormal] SigExpr
>   |  SigFun [SigFormal] SigExpr
>   |  SigApp SigExpr [SigActual]
>      deriving (Show)

Specification.

> data Spec
>   =  SpVal LVar Type
>   |  SpType LTyVar Type
>   |  SpData LCon
>   |  SpCon LCon Type
>   |  SpException Alt
>   |  SpModule LStrVar SigExpr
>   |  SpInterface LSigVar SigExpr
>      deriving (Show)

> type SigFormal
>   =  [LTyVar]
>
> type SigActual
>   =  [Type]

> instance Pretty Spec

> ppSpec (SpVal x t)      =  group (text "val " <> block 5 (ltext x <> text " :" <+> pType t))
> ppSpec (SpType x (TyFun fs t))
>                         =  group (text "type " <> block 5 (ppTyName x <> text " " <> strip (map pFormal fs) <> text " =" <+> pType t))
> ppSpec (SpType x t)     =  group (text "type " <> block 5 (ppTyName x <> text " =" <+> pType t))
> ppSpec (SpData x)       =  text "data " <> ppTyName x
> ppSpec (SpCon x (TyForall _ t))
>                         =  group (text "con " <> block 5 (ltext x <> text " :" <+> pType t))
> ppSpec (SpCon x t)      =  group (text "con " <> block 5 (ltext x <> text " :" <+> pType t))
> ppSpec (SpException (x, Nothing))
>                         =  text "exception " <> ltext x
> ppSpec (SpException (x, Just t))
>                         =  group (text "exception " <> block 5 (ltext x <+> pType t))
> ppSpec (SpModule x _)   =  text "module " <> ltext x  -- not showing signature
> ppSpec x                =  pretty x

> -- ppTyName = ltext -- print data types including identification stamps (debug)
> ppTyName = text . fst . splitTyName -- print data types without numbers

> splitTyName :: LCon -> (String, String)
> splitTyName (_,c) =
>   case break (=='#') c of
>     (d,_:n) -> (d,n)
>     _       -> (c,"")

> makeTyName :: Int -> String -> Con
> makeTyName n d = d ++ "#" ++ show n

%- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
\subsection{Abstract syntax: misc}
%- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 

> data Loc = At { line :: !Int, column :: !Int}
>
> instance Show Loc where
>   show (At l c) = "(line "  ++ show l ++ ", column " ++ show c ++ ")"
> instance Pretty Loc where
>   pretty (At 0 0) = empty
>   pretty (At l c) = parens (text "line " <> pretty l <> char ',' <> sp <> text "column " <> pretty c)

> unknown = At 0 0

> data Lab   
>   =  AlNumLab  String                         -- alphanumeric label
>   |  NumLab    Integer                        -- numeric label
>      deriving (Eq, Ord, Show)

> instance Pretty Lab

> ppLab (AlNumLab s) = text s
> ppLab (NumLab   n) = pretty n



